;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "aio" "1.1"
  "Async/await for Emacs Lisp."
  '((emacs "26.1"))
  :url "https://github.com/skeeto/emacs-aio"
  :commit "e0105f8ec8c47c12c0958442d0eaef72c94e8747"
  :revdesc "1.1-0-ge0105f8ec8c4"
  :authors '(("Christopher Wellons" . "wellons@nullprogram.com"))
  :maintainers '(("Christopher Wellons" . "wellons@nullprogram.com")))
